# coding: utf-8
from collections import OrderedDict
from collections import defaultdict

from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import FieldDoesNotExist
from django.db.transaction import atomic
from educommon.m3 import PackValidationMixin
from educommon.m3 import convert_validation_error_to
from educommon.m3 import get_id_value
from educommon.report.constructor.base import ModelDataSource
from educommon.report.constructor.config import report_constructor_config
from educommon.utils.ui import anchor100
from m3.actions.exceptions import ApplicationLogicException
from m3.actions.results import OperationResult
from m3.actions.results import PreJsonResult
from m3_ext.ui.results import ExtUIScriptResult
from objectpack.actions import BaseAction
from objectpack.actions import ObjectPack
from objectpack.exceptions import ValidationError
from objectpack.ui import BaseEditWindow
from objectpack.ui import make_combo_box

from ..base import ColumnDescriptor
from ..models import ReportColumn
from ..models import ReportTemplate
from ..registries import registry
from .ui import EditWindow
from .ui import ListWindow


class ColumnsAction(BaseAction):

    u"""Действие, предоставляющее данные для дерева колонок."""

    def context_declaration(self):
        return dict(
            data_source_name=dict(type='str'),
            parent_column_name=dict(type='str_or_none', default=None),
        )

    def run(self, request, context):
        if context.data_source_name not in registry:
            raise ApplicationLogicException(u'Не выбран источник данных.')

        if context.parent_column_name == u'-1':
            context.parent_column_name = None

        data_source = registry.get(context.data_source_name)

        if context.parent_column_name:
            parent_column = ColumnDescriptor.create(
                data_source.model, context.parent_column_name
            )
            columns = parent_column.get_nested_columns()
        else:
            columns = data_source.get_available_columns()

        return PreJsonResult(tuple(
            dict(
                name=column.accessor_name,
                full_name=column.full_accessor_name,
                title=column.title,
                leaf=not column.has_nested_columns(),
            )
            for column in columns
        ))


_BUILDERS_PACKAGE = 'educommon.report.constructor.builders'


class BuildAction(BaseAction):

    u"""Сборка отчета на основе указанного шаблона."""

    _builders = {
        ReportTemplate.EXCEL_SIMPLE:
            _BUILDERS_PACKAGE + '.excel.product.ReportBuilder',
        ReportTemplate.EXCEL_MERGED:
            _BUILDERS_PACKAGE + '.excel.with_merged_cells.ReportBuilder',
    }

    def context_declaration(self):
        return {
            self.parent.id_param_name: dict(type=int),
            'format': dict(type='int_or_none', default=None),
        }

    def _get_report_template(self, request, context):
        # Загрузка шаблона
        try:
            return ReportTemplate.objects.get(
                pk=get_id_value(context, self.parent)
            )
        except ReportTemplate.DoesNotExist:
            raise ApplicationLogicException(self.parent.MSG_DOESNOTEXISTS)

    def _check_report_template(self, report_template):
        # Проверка наличия в системе источника данных с именем, указанным в
        # шаблоне.
        if report_template.data_source_name not in registry:
            raise ApplicationLogicException(
                u'Источник данных {} не существует.'
                .format(report_template.data_source_name)
            )

        data_source = registry.get(report_template.data_source_name)

        # Проверка доступности данных для всех колонок шаблона отчета.
        try:
            for report_column in report_template.columns.all():
                data_source.get_column_descriptor(report_column.name)
        except FieldDoesNotExist as error:
            raise ApplicationLogicException(
                u'Колонка {} недоступна'.format(unicode(error))
            )

    def _check_params(self, request, context, report_template):
        u"""Возвращает True, если параметры сборки указаны верно."""
        if context.format is None:
            return report_template.format in self._builders

        if (
            report_template.format == ReportTemplate.USER_DEFINED and
            context.format in self._builders
        ):
            return True

        return False

    def _get_params_window(self, request, context, report_template):
        win = BaseEditWindow()

        win.field__format = make_combo_box(
            label=u'Формат отчета',
            name='format',
            data=ReportTemplate.FORMAT_CHOICES[1:],
        )

        win.form.items.extend(anchor100(
            win.field__format,
        ))

        win.set_params(dict(
            form_url=self.get_absolute_url(),
            height=100,
            title=u'Параметры сборки',
        ))

        return ExtUIScriptResult(win, context)

    def _check_format(self, format):
        if format not in self._builders:
            raise ApplicationLogicException(u'Указан неверный формат отчета.')

    def _build_report(self, request, context, report_template):
        report_template_id = get_id_value(context, self.parent)
        current_user_func = report_constructor_config.current_user_func
        task_class = report_constructor_config.async_task
        user = current_user_func(request)
        content_type = ContentType.objects.get_for_model(user)
        params = {
            'object_id': user.id,
            'content_type': content_type,
            'report_template_id': report_template_id,
            'format': context.format or None
        }
        task_class().apply_async(None, params)

        return OperationResult(
            message=u'Внимание! Задача поставлена в очередь! Результаты '
                    u'будут доступны в реестре Асинхронных задач.'
        )

    def run(self, request, context):
        report_template = self._get_report_template(request, context)
        self._check_report_template(report_template)
        if context.format:
            self._check_format(context.format)

        if self._check_params(request, context, report_template):
            return self._build_report(request, context, report_template)
        else:
            return self._get_params_window(request, context, report_template)


class Pack(PackValidationMixin, ObjectPack):

    u"""Пак реестра шаблонов отчетов."""

    model = ReportTemplate
    _is_primary_for_model = False

    columns = (
        dict(
            data_index='title',
            header=getattr(model, '_meta').get_field('title').verbose_name,
        ),
    )
    list_sort_order = ('title',)

    list_window = ListWindow
    add_window = edit_window = EditWindow

    def __init__(self):
        super(Pack, self).__init__()
        # ---------------------------------------------------------------------
        self.columns_action = ColumnsAction()
        self.build_action = BuildAction()

        self.actions.extend((
            self.columns_action,
            self.build_action,
        ))
        # ---------------------------------------------------------------------

    def declare_context(self, action):
        result = super(Pack, self).declare_context(action)

        if action is self.save_action:
            result.update(
                title=dict(type='unicode'),
                data_source_name=dict(type='str'),
                columns=dict(type='json'),
            )

        return result

    def get_list_window_params(self, params, request, context):
        result = super(Pack, self).get_list_window_params(
            params, request, context
        )

        result['build_action_url'] = self.build_action.get_absolute_url()

        return result

    def _get_columns(self, report_template):
        data_source = registry.get(report_template.data_source_name)

        descriptors = OrderedDict()
        report_columns = OrderedDict(
            (report_column.name, report_column)
            for report_column in report_template.columns.order_by('index')
        )
        for report_column in report_columns.itervalues():
            descriptor = data_source.get_column_descriptor(report_column.name)
            while descriptor:
                if descriptor.full_accessor_name not in descriptors:
                    descriptors[descriptor.full_accessor_name] = descriptor
                descriptor = descriptor.parent

        # Формирование словаря вложенных колонок
        hierarchy = defaultdict(list)
        for descriptor in descriptors.itervalues():
            if not descriptor.is_root():
                hierarchy[descriptor.parent.full_accessor_name].append(
                    descriptor
                )

        # Формирование итогового описания колонок шаблона отчета.
        def get_data(descriptor):
            result = dict(
                name=descriptor.accessor_name,
                full_name=descriptor.full_accessor_name,
                title=descriptor.title,
                leaf=not descriptor.has_nested_columns(),
                nested=tuple(
                    get_data(descriptor)
                    for descriptor in hierarchy[descriptor.full_accessor_name]
                ),
            )

            if descriptor.accessor_name in report_columns:
                report_column = report_columns[descriptor.accessor_name]
                result['overridden_title'] = report_column.title

            return result

        return tuple(
            get_data(descriptor)
            for descriptor in descriptors.itervalues()
            if descriptor.is_root()
        )

    def get_edit_window_params(self, params, request, context):
        result = super(Pack, self).get_edit_window_params(
            params, request, context
        )

        result['available_columns_action_url'] = (
            self.columns_action.get_absolute_url()
        )

        result['data_sources'] = registry.iteritems()

        if params['create_new']:
            result['columns'] = ()
        else:
            result['columns'] = self._get_columns(params['object'])

        return result

    @convert_validation_error_to(ValidationError)
    @atomic
    def save_row(self, report_template, create_new, request, context):
        if context.data_source_name not in registry:
            raise ApplicationLogicException(
                u'Источник данных "{}" не существует.'
                .format(context.data_source_name)
            )

        super(Pack, self).save_row(
            report_template, create_new, request, context
        )

        # Проверка наличия в источнике данных указанных колонок
        data_source = ModelDataSource(context.data_source_name)
        column_descriptors = []
        for column_params in context.columns:
            if 'accessor_name' not in column_params:
                raise ApplicationLogicException(
                    u'Колонки отчета заданы неверно.'
                )

            try:
                column_descriptors.append(ColumnDescriptor.create(
                    model=data_source.model,
                    field_name=column_params['accessor_name'],
                ))
            except FieldDoesNotExist:
                raise ApplicationLogicException(
                    u'Колонки "{}" нет в источнике данных "{}".'
                    .format(column_params['accessor_name'], data_source.title)
                )
        # ---------------------------------------------------------------------
        # Сохранение колонок.

        report_columns = OrderedDict(
            (report_column.name, report_column)
            for report_column in report_template.columns.order_by('index')
        )

        # Сначала нужно удалить колонки, которых нет в запросе.
        request_columns = {
            column_descriptor.full_accessor_name
            for column_descriptor in column_descriptors
        }
        for accessor_name in set(report_columns) - request_columns:
            report_column = report_columns.pop(accessor_name)
            report_column.delete()

        created_report_columns = []
        for index, column_descriptor in enumerate(column_descriptors, 1):
            report_column = report_columns.get(
                column_descriptor.full_accessor_name
            )

            if report_column is None:
                created_report_columns.append(
                    ReportColumn.objects.create(
                        report_template=report_template,
                        name=column_descriptor.full_accessor_name,
                        index=index,
                        title=column_descriptor.title,
                    )
                )

            elif (
                report_column.name != column_descriptor.full_accessor_name or
                report_column.index != index or
                report_column.title != column_descriptor.title
            ):
                report_column.name = column_descriptor.full_accessor_name
                report_column.index = index
                report_column.title = column_descriptor.title

                report_column.clean_and_save()

    def extend_menu(self, menu):
        return menu.SubMenu(
            u'Отчеты',
            menu.SubMenu(
                u'Конструктор отчетов',
                menu.Item(u'Редактор шаблонов', self.list_window_action),
            )
        )
